import { BaseConnection, WsConnection } from "tsrpc";
import { server } from "../..";
import { RandomEx } from "../../Engine/Utils/Number/RandomEx";
import { EGameState } from "../../shared/protocols/define/enum";
import { MsgChangeState } from "../../shared/protocols/room/MsgChangeState";
import { MsgGoToGame } from "../../shared/protocols/room/MsgGoToGame";
import { ServiceType } from "../../shared/protocols/serviceProto";
import { sleep } from "../../utils";
import { RoomConnData } from "./RoomUtils";


/**
 * Room
 */
export default class Room {

	//#region static

	private static maxRoomId: number = 0;
	private static rooms: { [roomId: number]: Room } = {};

	//#endregion

	//#region public

	public RoomId: number = 0;

	//#endregion

	//#region private

	private conns: BaseConnection[] = [];

	private connData: Map<BaseConnection<any>, RoomConnData> = new Map<BaseConnection, RoomConnData>();

	private type: number;

	//#endregion

	//#region Lifecycle

	/**
	 *
	 */
	constructor(type: number) {
		// 给每个新房间生成一个唯一的 ID
		this.RoomId = ++Room.maxRoomId;
		Room.rooms[this.RoomId] = this;
		this.type = type;
	}

	//#endregion

	//#region Custom

	/** GetRoom */
	public static GetRoom(roomId: number): Room {
		return this.rooms[roomId];
	}

	//#endregion

	//#region API

	/** Exit */
	public Exit(conn: BaseConnection): void {
		this.conns.splice(this.conns.indexOf(conn), 1);
		server.broadcastMsg("room/Exit", 0, <WsConnection<ServiceType>[]>this.conns);
		Room.rooms[this.RoomId] = null;
		delete Room.rooms[this.RoomId];
	}

	/** GotoGame */
	public GotoGame(): void {
		let data: MsgGoToGame = [this.type, this.conns as any];
		server.broadcastMsg("room/GoToGame", data, <WsConnection<ServiceType>[]>this.conns);
	}

	/**
	 * ChangeState
	 * @param needData true就是要帶connData.answer
	 * @param state 沒帶就是使用connData.state ?? EGameState.Ready
	 * @param conn 沒帶就是會+apiNum
	 */
	public ChangeState(needData: boolean = false, state?: EGameState, conn?: BaseConnection): void {
		this.connData.forEach((connData: RoomConnData, key: BaseConnection) => {
			let target: WsConnection<ServiceType>[] = [];
			let data: MsgChangeState = [];
			if (conn) {
				if (state) {
					target = <WsConnection<ServiceType>[]>this.conns.filter((filterConn) => filterConn === key);
					const apiNum: number = conn === key ? 100 : 200;
					data = [state + apiNum];
				} else if (conn) {
					target = <WsConnection<ServiceType>[]>this.conns.filter((filterConn) => filterConn === key);
					const apiNum: number = conn === key ? 100 : 200;
					data = [(connData.state ?? EGameState.Ready) + apiNum];
				}
			} else {
				if (state) {
					target = <WsConnection<ServiceType>[]>this.conns.filter((filterConn) => filterConn === key);
					data = [state];
				} else {
					target = <WsConnection<ServiceType>[]>this.conns.filter((filterConn) => filterConn === key);
					data = [connData.state];
				}
			}
			if (needData) {
				data.push(connData.answer);
			}
			server.broadcastMsg("room/ChangeState", data, target);
		});
		if (conn) {
			this.setState(conn, state);
		} else {
			this.conns.forEach((conn: BaseConnection) => {
				this.setState(conn, state);
			});
		}
		switch (state) {
			case EGameState.Ready:
				this.Ready(conn);
				break;

			case EGameState.TurnEnd:
				this.TurnEnd(conn);
				break;

			case EGameState.WaitNewGame: {
				let canStart: boolean = true;
				this.connData.forEach((connData: RoomConnData) => {
					if (connData.state !== EGameState.WaitNewGame) {
						canStart = false;
					}
				});
				if (canStart) {
					this.conns.forEach((conn: BaseConnection) => {
						this.setState(conn, EGameState.Ready);
					});
					this.Ready(conn);
				}
				break;
			}

			default:
				break;
		}
	}

	/** SendAns */
	public SendAns(conn: BaseConnection, answer: number): boolean {
		const index: number = this.conns.indexOf(conn);
		const nextIndex: number = (index + 1) % this.conns.length;
		const nextConn: BaseConnection = this.conns[nextIndex];
		const nextData: RoomConnData = this.connData.get(nextConn);
		const nextAns: number = nextData.answer;
		const data: boolean = nextAns === answer;
		if (data) {
			this.GameOver();
		}
		return data;
	}

	//#endregion

	//#region ChangeState

	/** Ready */
	public Ready(conn: BaseConnection): void {
		const self: this = this;
		const cardCount: number = 24;
		this.setState(conn, EGameState.Ready);
		let canStart: boolean = true;
		this.connData.forEach((connData: RoomConnData) => {
			if (connData.state === undefined || connData.state !== EGameState.Ready) {
				canStart = false;
			}
		});
		if (canStart) {
			const now: number = RandomEx.GetInt(0, 2);
			this.connData.forEach((connData: RoomConnData, key: BaseConnection) => {
				const card: number = RandomEx.GetInt(0, cardCount);
				connData.state = key === self.conns[now] ? EGameState.MyTurnCards : EGameState.MyWait;
				connData.answer = card;
				self.connData.set(key, connData);
			});
			this.ChangeState(true);
		}
	}

	/** TurnEnd */
	public TurnEnd(conn: BaseConnection): void {
		const self: this = this;
		this.connData.forEach((connData: RoomConnData, key: BaseConnection) => {
			connData.state = EGameState.MyWait;
			self.connData.set(key, connData);
		});
		const index: number = this.conns.indexOf(conn);
		const nextIndex: number = (index + 1) % this.conns.length;
		const nextConn: BaseConnection = this.conns[nextIndex];
		const nextData: RoomConnData = this.connData.get(nextConn);
		nextData.state = EGameState.MyTurnCards;
		this.connData.set(nextConn, nextData);
		this.ChangeState();
	}

	//#endregion

	//#region Custom

	/** setState */
	private setState(conn: BaseConnection, state: EGameState): void {
		const data: RoomConnData = this.connData.get(conn);
		data.state = state;
		this.connData.set(conn, data);
	}

	/** Join */
	public Join(conn: BaseConnection): number {
		this.conns.push(conn);
		const data: RoomConnData = {
			state: undefined,
			answer: undefined,
		};
		this.connData.set(conn, data);
		return this.type;
	}

	/** GameOver */
	public async GameOver(): Promise<void> {
		await sleep(500);
		this.ChangeState(false, EGameState.GameOver);
	}

	/** RestartGame */
	public RestartGame(): void {

	}

	/** ConnCount */
	public ConnCount(): number {
		return this.conns.length;
	}

	//#endregion
}
